"""Useful functions for defining loss functions, metrics, model loading and data augmentation"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/utils/01_helper_func.ipynb.

# %% auto 0
__all__ = ['load_module', 'AbstractModel']

# %% ../../nbs/utils/01_helper_func.ipynb 3
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import xarray as xr

import importlib.util as import_utils
from pathlib import Path
from abc import ABC, abstractmethod

# %% ../../nbs/utils/01_helper_func.ipynb 4
def load_module(
  module_name: str, # The name of the module to load
  module_path: Path, # The path to the module
) -> object: # The loaded module
    "Load a module from a file path."
    spec = import_utils.spec_from_file_location(self._FUNCTION_MODULE_NAME.replace(".py", ""), self.model_path / self._FUNCTION_MODULE_NAME)
    module = import_utilsmodule_from_spec(spec)
    spec.loader.exec_module(module)
    return module

# %% ../../nbs/utils/01_helper_func.ipynb 5
class AbstractModel(ABC):
  
  @abstractmethod
  def predict(self, x: xr.Dataset) -> xr.Dataset:
    pass
  
  @abstractmethod
  def contributions(self, x: xr.Dataset) -> xr.Dataset:
    pass
  
  def __repr__(self) -> str:
    return f"Model"



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/budget_optimizer',
                'doc_host': 'https://optimizer.mattreda.pro',
                'git_url': 'https://github.com/redam94/budget_optimizer',
                'lib_path': 'budget_optimizer'},
  'syms': { 'budget_optimizer.optimizer': { 'budget_optimizer.optimizer.Optimizer': ( 'optimizer.html#optimizer',
                                                                                      'budget_optimizer/optimizer.py'),
                                            'budget_optimizer.optimizer.Optimizer.__init__': ( 'optimizer.html#optimizer.__init__',
                                                                                               'budget_optimizer/optimizer.py'),
                                            'budget_optimizer.optimizer.Optimizer._load_config': ( 'optimizer.html#optimizer._load_config',
                                                                                                   'budget_optimizer/optimizer.py'),
                                            'budget_optimizer.optimizer.Optimizer._load_loss_fn': ( 'optimizer.html#optimizer._load_loss_fn',
                                                                                                    'budget_optimizer/optimizer.py'),
                                            'budget_optimizer.optimizer.Optimizer._optimizer_array_to_budget': ( 'optimizer.html#optimizer._optimizer_array_to_budget',
                                                                                                                 'budget_optimizer/optimizer.py'),
                                            'budget_optimizer.optimizer.Optimizer.optimize': ( 'optimizer.html#optimizer.optimize',
                                                                                               'budget_optimizer/optimizer.py'),
                                            'budget_optimizer.optimizer.Optimizer.reload_config': ( 'optimizer.html#optimizer.reload_config',
                                                                                                    'budget_optimizer/optimizer.py')},
            'budget_optimizer.utils.funcs': { 'budget_optimizer.utils.funcs.AbstractModel': ( 'utils/helper_func.html#abstractmodel',
                                                                                              'budget_optimizer/utils/funcs.py'),
                                              'budget_optimizer.utils.funcs.AbstractModel.__repr__': ( 'utils/helper_func.html#abstractmodel.__repr__',
                                                                                                       'budget_optimizer/utils/funcs.py'),
                                              'budget_optimizer.utils.funcs.AbstractModel.contributions': ( 'utils/helper_func.html#abstractmodel.contributions',
                                                                                                            'budget_optimizer/utils/funcs.py'),
                                              'budget_optimizer.utils.funcs.AbstractModel.predict': ( 'utils/helper_func.html#abstractmodel.predict',
                                                                                                      'budget_optimizer/utils/funcs.py'),
                                              'budget_optimizer.utils.funcs.load_module': ( 'utils/helper_func.html#load_module',
                                                                                            'budget_optimizer/utils/funcs.py')},
            'budget_optimizer.utils.model_classes': { 'budget_optimizer.utils.model_classes.BaseBudgetModel': ( 'utils/model_classes.html#basebudgetmodel',
                                                                                                                'budget_optimizer/utils/model_classes.py'),
                                                      'budget_optimizer.utils.model_classes.BaseBudgetModel.__init__': ( 'utils/model_classes.html#basebudgetmodel.__init__',
                                                                                                                         'budget_optimizer/utils/model_classes.py'),
                                                      'budget_optimizer.utils.model_classes.BaseBudgetModel._get_budget_to_data': ( 'utils/model_classes.html#basebudgetmodel._get_budget_to_data',
                                                                                                                                    'budget_optimizer/utils/model_classes.py'),
                                                      'budget_optimizer.utils.model_classes.BaseBudgetModel._get_model_loader': ( 'utils/model_classes.html#basebudgetmodel._get_model_loader',
                                                                                                                                  'budget_optimizer/utils/model_classes.py'),
                                                      'budget_optimizer.utils.model_classes.BaseBudgetModel.contributions': ( 'utils/model_classes.html#basebudgetmodel.contributions',
                                                                                                                              'budget_optimizer/utils/model_classes.py'),
                                                      'budget_optimizer.utils.model_classes.BaseBudgetModel.predict': ( 'utils/model_classes.html#basebudgetmodel.predict',
                                                                                                                        'budget_optimizer/utils/model_classes.py')},
            'budget_optimizer.utils.model_helpers': { 'budget_optimizer.utils.model_helpers.AbstractModel': ( 'utils/model_helpers.html#abstractmodel',
                                                                                                              'budget_optimizer/utils/model_helpers.py'),
                                                      'budget_optimizer.utils.model_helpers.AbstractModel.__repr__': ( 'utils/model_helpers.html#abstractmodel.__repr__',
                                                                                                                       'budget_optimizer/utils/model_helpers.py'),
                                                      'budget_optimizer.utils.model_helpers.AbstractModel.contributions': ( 'utils/model_helpers.html#abstractmodel.contributions',
                                                                                                                            'budget_optimizer/utils/model_helpers.py'),
                                                      'budget_optimizer.utils.model_helpers.AbstractModel.predict': ( 'utils/model_helpers.html#abstractmodel.predict',
                                                                                                                      'budget_optimizer/utils/model_helpers.py'),
                                                      'budget_optimizer.utils.model_helpers.load_module': ( 'utils/model_helpers.html#load_module',
                                                                                                            'budget_optimizer/utils/model_helpers.py'),
                                                      'budget_optimizer.utils.model_helpers.load_yaml': ( 'utils/model_helpers.html#load_yaml',
                                                                                                          'budget_optimizer/utils/model_helpers.py')}}}



"""Optimizers for models"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_optimizer.ipynb.

# %% auto 0
__all__ = ['Optimizer']

# %% ../nbs/00_optimizer.ipynb 5
import pyswarms as ps
import numpy as np
import pandas as pd

from .utils.model_classes import BaseBudgetModel
from budget_optimizer.utils.model_helpers import (
  load_module,
  load_yaml,
  BudgetType, 
  AbstractModel
)

from pathlib import Path

# %% ../nbs/00_optimizer.ipynb 6
class Optimizer:
    """Optimizer wrapper for the pyswarms package"""
    _CONFIG_YAML = 'optimizer_config.yaml'
    _MODULE_FILE = "optimizer_config.py"
    
    def __init__(
        self, 
        model: BaseBudgetModel, # The model to optimize
        config_path: str|Path # Path to the configuration files
        ):
        
        self.model: AbstractModel = model
        self._config_path: Path = Path(config_path) if isinstance(config_path, str) else config_path
        self._optimal_budget: BudgetType = None
        self._optimal_prediction: xr.DataArray = None
        self._optimal_contribution: xr.Dataset = None
        self._config = self._load_config()
        self._loss_fn = self._load_loss_fn()
        
    def _load_config(self):
        config = load_yaml(self._config_path / self._CONFIG_YAML)
        return config
    
    def reload_config(self):
        self._config = self._load_config()
        return self
    
    def _load_loss_fn(self):
        """Load the loss function from the config file"""
        module = load_module(self._MODULE_FILE.replace(".py", ""), self._config_path / self._MODULE_FILE)
        return module.loss_fn
    
    def _optimizer_array_to_budget(self, array: np.ndarray) -> BudgetType:
        """Convert the optimizer array to a budget"""
        initial_budget: BudgetType = self._config['initial_budget']
        budget: BudgetType = {}
        
        for i, key in enumerate(initial_budget.keys()):
            budget[key] = array[i]
        return budget
    
    
    def optimize(self, n_particles: int = 10, n_iterations: int = 100):
        """Optimize the model"""
        optimizer = ps.single.GlobalBestPSO(n_particles=n_particles, dimensions=self._config['dimensions'])
        optimizer.optimize(self._loss_fn, iters=n_iterations)
        self._optimal_budget = optimizer.pos_best
        self._optimal_prediction = self.model.predict() # The optimizer minimizes the cost, so we need to negate it
        self._optimal_contribution = self.model.get_contribution(self._optimal_budget)
        return self
